import re

import lang

# Match the name of a dumpfile
dumpfile_name = re.compile(r'(?P<prefix>.*?)wiki-(?P<date>\d{8})-pages-articles.xml')

intrawiki_link = re.compile(r"\[\[(?P<target>.*?\|)?(?P<anchor>.*?)\]\]")

lang_prefixes = '|'.join(lang.prefixes)
lang_link = re.compile(r"\[\[(?P<prefix>"+lang_prefixes+"):(?P<title>.+?)\]\]")

category_keywords = '|'.join(lang.category_identifier.values())
category_link = re.compile(r"\[\[(?P<title>("+category_keywords+"):(?P<category>.+?))\|(.+?)\]\]")
category_name = re.compile(r"(?P<keyword>("+category_keywords+")):(?P<category>.+)")

redirect = re.compile(r"\#REDIRECT (?P<target>.*)")

template = re.compile(r"\{\{(?P<target>.*?)\}\}")



import os
import logging
import time
import shelve

from config import config

scratchpath = config.get('paths','scratch')

class shelved(object):
  """ Decorator maker for shelved """
  def __init__(self, path):
    self.path = os.path.join(config.get('paths','scratch'), path)
    
  def __call__(self, func):
    return Shelved(self.path, func)

class Shelved(object):
   """Decorator that caches a function's return value each time it is called.
   If called later with the same arguments, the cached value is returned, and
   not re-evaluated. Also shelves the result in the specified file for future
   use.
   """
   def __init__(self, shelf_path, func):
      self.func = func
      self.shelf = shelve.open(shelf_path)
      self.logger = logging.getLogger('wikidump.common.Shelved')

   def __del__(self):
      self.shelf.close() 

   def __call__(self, *args):
      key = str(args)
      try:
        return self.shelf[key]
      except KeyError:
        self.logger.debug("Not in cache: %s", key)
        self.shelf[key] = value = self.func(*args)
        return value
      except TypeError:
        # uncachable -- for instance, passing a list as an argument.
        # Better to not cache than to blow up entirely.
        self.logger.warning("Failed to cache")
        return self.func(*args)
   def __repr__(self):
      """Return the function's docstring."""
      return self.func.__doc__




"""
Model of a wikipedia dump
"""

import shelve 
import hashlib
import os
import re
import logging
import xml.etree.ElementTree as etree

import regexps
import utils
from config import config
from parser import PageOffsetParser


class Dump:
  logger = logging.getLogger('wikidump.model.Dump')

  def _open_shelf(self, name):
    "Open a shelf-file in a pre-determined location by name"
    path = os.path.join(self.cache_path, name)
    return shelve.open(path)

  def __init__(self, xml_path, build_index=False):
    self.xml_path = os.path.abspath(xml_path)
    self.xml_file = open(self.xml_path)
    # May want to hash the file instead for portability
    path_hash = hashlib.sha1(os.path.basename(self.xml_path)).hexdigest()
    self.cache_path = os.path.join(config.get('paths','scratch'), path_hash)

    self.logger.info("============================================")
    self.logger.info("Loading data for %s", self.xml_path)

    if not os.path.exists(self.cache_path):
      self.logger.info("Creating %s", self.cache_path)
      os.mkdir(self.cache_path)

    # Open a metadata shelf
    self.metadata = self._open_shelf('metadata')

    # Compute size of dump
    try:
      size = self.metadata['size']
    except KeyError:
      size = int(os.popen("grep '<page>' %s | wc -l" % self.xml_path).read())
      self.metadata['size'] = size
      self.logger.info("Found %d pages", size)
    self.logger.debug("Size: %d pages", size)

    # Mapping from page index to position in file
    self.page_offsets = self._open_shelf('page_offsets')
    #self.logger.debug("Currently know of %d page offsets", len(self.page_offsets))

    if build_index and len(self.page_offsets) < size:
      self.logger.info("Calculating page offsets")
      parser = PageOffsetParser(open(self.xml_path), self.page_offsets) 
      parser.run()
      self.page_offsets.sync()
    else:
      self.logger.debug("Not calculating page offsets")

    # Mapping from page title to page index
    self.page_titles = self._open_shelf('page_titles')
    #self.logger.debug("Currently know of %d page titles", len(self.page_titles))
    if build_index and len(self.page_titles) < size:
      self.logger.info("Building page title index")
      # Skip forwards if we ended up with a partial index due to an aborted 
      # build
      for i in range(len(self.page_titles),size):
        tree = etree.fromstring(self.get_raw(i))
        # Need to encode as etree will return both str and unicode
        title = tree.find('title').text.encode('utf8')
        if title in self.page_titles:
          self.logger.warning("Already had '%s', index %d", title, self.page_titles[title])
        else:
          self.page_titles[title] = i
    else:
      self.logger.debug("Not building page title index")

    # Mapping from categories to page indices 
    self.categories = self._open_shelf('categories')
    if build_index and 'categories_mapped' not in self.metadata: 
      self.logger.info("Building category index")
      cat_map = utils.category_map(self)
      # Copy into shelf
      for key in cat_map:
        self.categories[key] = cat_map[key]
      self.metadata['categories_mapped'] = True
    else:
      self.logger.debug("Not building category index")

    self.page_lengths = self._open_shelf('page_lengths')
    self.logger.debug("__init__ complete")

  def get_raw(self, index):
    "Get raw xml dump data for a given index"
    f = self.xml_file

    start_offset = self.page_offsets[str(index+1)]
    f.seek(start_offset)
    try:
      end_offset = self.page_offsets[str(index+2)]
      return f.read(end_offset - start_offset)
    except KeyError:
      # Handle the corner case which is the very last entry
      raw = f.read()
      match = re.search(r'.*</page>', raw)
      return raw[:match.end()]
      
  def get_page_index(self, title):
    "Look up the index of a page based on its title"
    return self.page_titles[title] 

  def get_page_contents(self, index):
    "Get the full text of a page by index"
    tree = etree.fromstring(self.get_raw(index))
    return tree.find('revision').find('text').text.encode('utf8')

  def get_page_length(self, index):
    "Look up the length of a page given an index"
    try:
      return self.page_lengths[str(index)]
    except KeyError:
      length = len(self.get_page_contents(index))
      self.page_lengths[str(index)] = length
      return length

  def get_page_contents_by_title(self, title):
    "Get the contents of a page with a given title"
    index = self.get_page_index(title)
    return self.get_page_contents(index)

  def get_page(self, title):
    "Return a Page object for the page of the given title"
    index = self.get_page_index(title)
    return Page(self.get_raw(index))

  def get_page_by_index(self, index):
    "Return a Page object given the raw index of the page"
    return Page(self.get_raw(index))

  def get_dumpfile_prefix(self):
    "Return the prefix code associated with the filename"
    filename = os.path.basename(self.xml_path)
    return regexps.dumpfile_name.match(filename).groups()[0]

class Page:
  """
  Represents a single page in a wikidump
  """
  logger = logging.getLogger('wikidump.model.Page')

  def __init__(self, string):
    """
    @param string: A string containin the raw xml version of the page
    """
    self.dom = etree.fromstring(string)
    self.text = self.dom.find('revision').find('text').text
    self.title = self.dom.find('title').text.encode('utf8')

    try:
      self.text = self.text.encode('utf8')
    except AttributeError:
      #self.logger.warning("Empty text!")
      self.text = ''

  def lang_equiv(self, prefix):
    """ Returns the page title for the equivalent page in the given language prefix
    """
    lang_links = dict(regexps.lang_link.findall(self.text))
    try:
      return lang_links[prefix]
    except KeyError:
      return None

  def categories(self):
    """ Returns the set of categories this page is member of.
        changed from 0.1 : was previously a list. Do not want duplicates.
    """
    try: 
      return set(zip(*regexps.category_link.findall(self.text))[0])
    except IndexError: #No categories
      return []




#TODO FIX ME UP!!!
import numpy
import random
import logging

from cPickle import dump

from utils import dumpSize, all_prefixes, page_parser
from parser import strip_mediawiki_markup

logger = logging.getLogger('wikidump.segment')

def create_dataset( num_doc
                  , distribution
                  , num_seg 
                  , text_transformer
                  , exclude = []
                  , seed = None
                  ):
  """
  Create a multilingual dataset from wikipedia data based on segments
  of monolingual documents.

  @param num_doc: number of documents in intended dataset
  @param distribution: Mapping from class name to proportion
  @param segments: Sequence of segments and their relative sizes
  @param exclude: Docids to be excluded from the final dataset
  """
  logger = logging.getLogger('wikidump.segment.create_dataset')
  logger.info('Creating dataset')
  docmap = {}
  classmap = {}
  # Set up a parser for each class
  parser = {}
  classlabels = []
  class_probs = numpy.empty( len(distribution), dtype=float) 
  dump_indices = [] 
  denominator = 0 
  logger.info('Setting up parsers')
  for i,classlabel in enumerate(distribution):
    parser[classlabel] = page_parser(classlabel)
    classlabels.append(classlabel)
    denominator += distribution[classlabel] 
    class_probs[i] = denominator 
    dumpsize = dumpSize(classlabel)
    indices = range(dumpsize)
    random.shuffle(indices)
    dump_indices.append(indices)
    logger.debug('Done for %s, %d indices', classlabel, dumpsize)
  logger.info('Parsers ready')

  # Normalize to a CDF
  class_probs /= denominator

  random.seed(seed)

  logger.info('Obtaining documents')
  for doc_num in range(num_doc):
    doc = ""
    classes = set() 
    segids = []
    for seg_num in range(num_seg):
      seg_class_index = numpy.searchsorted(class_probs, random.random())
      seg_class = classlabels[seg_class_index] 
      classes.add(seg_class)
      doc_index = dump_indices[seg_class_index].pop(0)
      segid = "_".join((seg_class, str(doc_index)))
      while segid in exclude:
        logger.debug('Ignored %s: in exclude', segid)
        try:
          doc_index = dump_indices[seg_class_index].pop(0)
        except IndexError:
          raise ValueError, "No more documents in class %s available" % classlabel 
        segid = "_".join((seg_class, str(doc_index)))
      segids.append(segid)
      content = text_transformer(parser[seg_class].get_entry(doc_index))
      seg_size = len(content)/num_seg
      seg_start = seg_size * seg_num
      seg_end = seg_start + seg_size
      doc += content[seg_start:seg_end] 
    docid = '-'.join(segids) 
    docmap[docid] = doc
    classmap[docid] = list(classes)
    logger.debug('Index: %d ID: %s', doc_num, docid)
  return docmap, classmap

if __name__ == "__main__":
  picklefile = open('segment.pickle', "w")
  distribution =  {}
  for p in all_prefixes:
    if len(p) > 3:
      logger.warning('Ignoring %s', p)
    else:
      distribution[p] = dumpSize(p)
  data = create_dataset( 5000 
                       , distribution 
                       , 2 
                       , strip_mediawiki_markup 
                       , exclude = []
                       , seed = 61383441363  
                       )
  dump(data, picklefile) 




import logging
from ConfigParser import SafeConfigParser

logger = logging.getLogger('wikidump.config')

default_config_path = 'wikidump.cfg'

config = SafeConfigParser()

if not config.read(default_config_path):
  logger.info("Did not find a configuration file. Creating one.")
  config.add_section('paths')
  config.set('paths', 'scratch', '/tmp')
  config.set('paths', 'xml_dumps', '/path/to/dumps')
  with open(default_config_path, 'w') as configfile:
    config.write(configfile)




from common.decorators import shelved
from utils import page_parser, dumpSize
from time import time
import os
from utils import all_prefixes, raw_doc_lengths
from doclength_threshold import docs_under_thresh
import numpy as n
from matplotlib.font_manager import FontProperties
import pylab as p
import operator
from utils import dumpSize

html_colors = [ 'aqua'
              , 'red'
              , 'blue'
              , 'fuchsia'
              , 'teal'
              , 'maroon'
              , 'purple'
              , 'navy'
              , 'green'
              , 'olive'
              , 'yellow'
              , 'lime'
              , 'gray'
              , 'silver'
              ]

def produce_graphs():
  length_stats_chart('output/wikipedia_doclength_stats_median.png', all_prefixes,1)
  length_stats_chart('output/wikipedia_doclength_stats_mean.png', all_prefixes,2)
  length_stats_chart('output/wikipedia_doclength_stats_sd.png', all_prefixes,3)
  #wiki_sizes_chart('output/wiki_size_chart.png', all_prefixes)  
  #wiki_sizes_chart('output/wiki_size_chart_10000_upper.png', all_prefixes, 10000)  
  #wiki_sizes_chart('output/wiki_size_chart_100000_upper.png', all_prefixes, 100000)  
  #wiki_proportions_chart('output/wiki_doclen_proportions.png', all_prefixes)

def overall_collection_stats():
  doc_lengths = []
  for prefix in all_prefixes:
    doc_lengths += raw_doc_lengths(prefix).values()
  print "All documents length mean: %.2f" % n.mean(doc_lengths)
  print "All documents length std: %.2f" % n.std(doc_lengths)
  print "All documents length median: %.2f" % n.median(doc_lengths)
  medlist, meanlist, stdlist = zip(*map(length_stats,all_prefixes))
  print "Mean of means: %.2f" % n.mean(meanlist)
  print "Mean of std: %.2f" % n.mean(stdlist)
  print "Mean of median: %.2f" % n.mean(medlist)

@shelved("length_stats")
def length_stats(prefix):
  doc_lengths = raw_doc_lengths(prefix).values()
  median = n.median(doc_lengths)
  mean   = n.mean(doc_lengths)
  std    = n.std(doc_lengths)
  return median, mean, std
    
def rounded_interval(min, max, points, divs=1):
  exact = (max - min) / points
  from math import log
  return (10 ** round(log(exact*divs,10))) / divs

def wiki_proportions_chart(path, prefixes):
  prefixes, sizes = zip(*sorted( [(pr, dumpSize(pr)) for pr in prefixes]
                               , key = operator.itemgetter(1)
                               )
                       )

  blockSize = 5 
  ind = p.arange(0, blockSize*len(prefixes), blockSize) # y location for groups
  height = 4 # bar height 

  #colors = ['g','r','c','m','y']
  thresholds = [5000, 2000,1000,500,200,100,50,20,10]
  colors = [str(float(i+1) / (len(thresholds)+1)) for i in xrange(len(thresholds))]
  colors.reverse()

  p.clf()
  """
  overall = p.barh( ind 
                  , [1.0] * len(ind) 
                  , height
                  , color = 'b'
                  , linewidth = 0
                  , align='center'
                  )
  """
  subbars = []
  for i, thresh in enumerate(thresholds) :
    subbars.append( p.barh( ind
                          , [ float(docs_under_thresh(pr, thresh)) / dumpSize(pr) for pr in prefixes]
                          , height
                          , color = colors[ i % len(colors) ] 
                          , linewidth = 0
                          , align='center'
                          )
                  )
  
  p.ylim(-height, len(prefixes) * blockSize)
  p.xlim(0, 1)
  yfontprop = FontProperties(size=4)
  xfontprop = FontProperties(size=4)
  p.xlabel('Proportion')
  p.ylabel('Language Code')
  p.title('Proportion of Documents Under Threshold')
  p.yticks(ind, prefixes, fontproperties = yfontprop)
  xmin, xmax = p.xlim()
  xtick_interval         = 0.1 
  p.xticks( p.arange(xmin,xmax,xtick_interval),fontproperties = xfontprop)
  p.gca().xaxis.grid(linestyle = '-', linewidth=0.15)
  p.gca().yaxis.tick_left()
  p.savefig(path, dpi=300)
  p.close()
  p.clf()
  
  
def wiki_sizes_chart(path, prefixes, upperlimit = None ):
  prefixes, sizes = zip(*sorted( [(pr, dumpSize(pr)) for pr in prefixes]
                               , key = operator.itemgetter(1)
                               )
                       )

  blockSize = 5 
  ind = p.arange(0, blockSize*len(prefixes), blockSize) # y location for groups
  height = 4 # bar height 

  #colors = ['g','r','c','m','y']
  colors = html_colors

  thresholds = [5000, 2000,1000,500,200,100,50,20,10]
  #colors = [str(float(i+1) / (len(thresholds)+1)) for i in xrange(len(thresholds))]
  #colors.reverse()

  overall = p.barh( ind 
                  , sizes
                  , height
                  , color = 'b'
                  , linewidth = 0
                  , align='center'
                  )
  subbars = []
  for i, thresh in enumerate(thresholds) :
    subbars.append( p.barh( ind
                          , [ docs_under_thresh(pr, thresh) for pr in prefixes]
                          , height
                          , color = colors[ i % len(colors) ] 
                          , linewidth = 0
                          , align='center'
                          )
                  )
  
  p.ylim(-height, len(prefixes) * blockSize)
  if upperlimit:
    p.xlim(0, upperlimit)
  yfontprop = FontProperties(size=4)
  xfontprop = FontProperties(size=4)
  p.xlabel('Documents')
  p.ylabel('Language Code')
  p.title('Number of Documents Under Threshold')
  p.yticks(ind, prefixes, fontproperties = yfontprop)
  xmin, xmax = p.xlim()
  xtick_interval         = rounded_interval(xmin, xmax, 20, 2) 
  p.xticks( p.arange(xmin,xmax,xtick_interval),fontproperties = xfontprop)
  p.gca().xaxis.grid(linestyle = '-', linewidth=0.15)
  p.gca().yaxis.tick_left()
  p.legend( [ b[0] for b in subbars]
          , map(str,thresholds)
          , prop = xfontprop
          , loc = 'lower right' 
          )


  p.savefig(path, dpi=300)
  p.close()
  p.clf()
  

def length_stats_chart(path, prefixes, sortby=1):
  stats = []
  for prefix in prefixes:
    med, m,s = length_stats(prefix)
    stats.append((prefix,med,m,s))

  stats.sort(key=operator.itemgetter(sortby))
  prefixes, med_list, mean_list, std_list = zip(*stats)

  blockSize = 8 
  ind = p.arange(0, blockSize*len(prefixes), blockSize) # y location for groups
  height = 3 # bar height 

  p3 = p.barh(ind, std_list, 2   * height, color = 'b', linewidth = 0)
  p2 = p.barh(ind, med_list, height, color = 'g', linewidth = 0)
  p1 = p.barh(ind+height, mean_list, height, color = 'r', linewidth = 0)
  
  p.ylim(-height, len(prefixes) * blockSize)
  yfontprop = FontProperties(size=4)
  xfontprop = FontProperties(size='smaller')
  p.xlabel('Unicode Codepoints')
  p.ylabel('Language Code')
  p.title('Descriptive Statistics for Document Lengths')
  p.gca().yaxis.tick_left()
  p.yticks(ind+height, prefixes, fontproperties = yfontprop)
  xmin, xmax = p.xlim()
  p.xticks( p.arange(xmin,xmax,1000),fontproperties = xfontprop)
  p.gca().xaxis.grid(linestyle = '-', linewidth=0.15)
  p.legend((p1[0], p2[0], p3[0]), ('Mean','Median','Standard Deviation'), prop = xfontprop, loc = 'lower right' )

  p.savefig(path, dpi=300)
  p.close()
  p.clf()

  

def doclength_histogram(path, prefix):
  values = p.array(raw_doc_lengths(prefix).values())
  num_bins = 1000
  bin_upper_limit = p.mean(values) + 3 * p.std(values)
  print "UL: "+ str(bin_upper_limit)
  bins = p.array(range(1,1001)) * (bin_upper_limit/1000.0)
  p.hist(values, bins)
  p.xlabel('Document size (unicode codepoints)')
  p.ylabel('Number of documents')
  p.title('Document Size Histogram for %s' % prefix)
  p.savefig(path, dpi=72)
  p.close()
  
def all_prefix_length_histograms(path):
  for prefix in all_prefixes:
    outpath = os.path.join(path, "%s_doclength_hist.png"%prefix)
    doclength_histogram(outpath, prefix)



import logging
from sgmllib import SGMLParser

logger = logging.getLogger('wikidump.parser')

class PageOffsetParser(SGMLParser):
  "Parser to compute offsets for <page> starts by document index"
  def __init__(self, input_file, mapping):
    input_file.seek(0)
    self.input_file = input_file
    self.mapping = mapping
    self.current_offset = 0
    self.count = 0
    SGMLParser.__init__(self)

  def start_page(self, attrs):
    self.count += 1
    self.mapping[str(self.count)] = self.current_offset

  def run(self):
    for line in self.input_file:
      self.feed(line)
      self.current_offset += len(line)



"""
Code useful for preparing datasets from wikidump data
"""
import re
import os
import shutil

import regexps
import utils

mediawiki_ignore_stringheads = '{}|=*#:'

def remove_mediawiki_syntax(text):
  "Remove non-content from a mediawiki page"
  text = regexps.redirect.sub('',text) # Remove redirects
  text = regexps.lang_link.sub('',text) # Remove language links
  text = regexps.template.sub('',text) # Remove templates TODO: Multi-line?
  text = regexps.intrawiki_link.sub(lambda x: x.group('anchor'), text) # Replace intrawiki links with anchor text
  text = re.sub('\n\n+', '\n\n', text) # Collapse multiple newlines
  return text.strip()

def paragraphs(text, minsize = 0):
  "Split text into paragraphs"
  return filter(lambda x: len(x) > minsize, text.split('\n\n'))

def lang_dist(langs):
  "mapping of the distribution of documents in a given set of languages"
  dumps = utils.load_dumps(langs)
  dist = dict((d.get_dumpfile_prefix(), d.metadata['size']) for d in dumps.values())
  return dist

try:
  import sampling
  from counter import Counter

  @sampling.CheckRNG
  def lang_sample(size, langs, rng=None):
    background = lang_dist(langs) 
    dist_raw = sampling.weighted_choice\
                 ( background
                 , count=size
                 , rng=rng
                 )

    dist = Counter(dist_raw)
    return dist 
except ImportError:
  # Skip if the sampling import fails - this will be because the numpy import in sampling failed
  pass

def refresh_dir(path):
  if os.path.exists(path) and not os.path.isdir(path):
    raise ValueError, "%s is not a directory" % path
  if os.path.exists(path):
    shutil.rmtree(path)
  os.mkdir(path)
  return path





"""
Provides numpy-based routines for performing sampling tasks
"""
import numpy
import math

allow_default_rng = True
default_rng = numpy.random.mtrand.RandomState()

class RNGError(Exception): pass

class CheckRNG(object):
  """ Decorator that checks that an RNG has been provided, or provides a default one 
      if it is allowed by the module's current setting, as controlled by the 
      'allow_default_rng' variable
  """
  def __init__(self, func):
    self.func = func

  def __call__(self, *args, **kwargs):
    if 'rng' not in kwargs or kwargs['rng'] is None:
      if allow_default_rng:
        kwargs['rng'] = default_rng
      else:
        raise RNGError, "Default RNG disabled - please pass an explicit RNG"
    return self.func(*args, **kwargs)

  def __repr__(self):
    return repr(self.func)

  def __str__(self):
    return str(self.func)

@CheckRNG
def weighted_choice(items, weights=None, count=1, rng=None):
  """ Weighted random choice of n items
  """
  if hasattr(items, 'values'): 
    items, weights = zip(*items.items())
   
  if not weights: weights = len(items) * [1]

  # Normalize to 1 if required
  t = float(sum(weights))

  # Uniform chance if weights sum to 0
  if t == 0.0: weights = [ 1/len(items) for i in items ]
  elif t != 1.0: weights = [ w/t for w in weights ]

  dist = numpy.cumsum(weights)
  result = numpy.searchsorted(dist, rng.random_sample(count))

  if count == 1:
    return list(numpy.array(items)[result])[0]
  else:
    return list(numpy.array(items)[result])

@CheckRNG
def partition(num_items, weights, probabilistic = False, rng = None):
  """
  Partition a number of items into partitions according to weights
  Modified from hydrat: http://hydrat.googlecode.com/svn/trunk/task/sampler.py
  @return: map of doc_index -> partition membership
  @rtype: numpy boolean array (items * partitions)
  """
  weights = numpy.array(weights)
  probabilities = weights / float(numpy.sum(weights))

  num_parts   = len(probabilities)
  partition_map = numpy.zeros((num_items,num_parts), dtype = 'bool')
  if probabilistic:
    c = numpy.cumsum(probabilities)
    r = rng.random_sample(num_items)
    partition_indices = numpy.searchsorted(c, r)
    for doc_index, part_index in enumerate(partition_indices):
      partition_map[doc_index, part_index] = True
  else:
    partition_sizes = numpy.floor(probabilities * num_items).astype(int)
    items_partitioned = partition_sizes.sum()
    gap = num_items - items_partitioned
    # Distribute the gap amongst the biggest partitions, adding one to each
    # This behaviour was deprecated as it does not behave well when partitions
    # all have the same size. The later partitions get favored by nature of the
    # argsort.
    # distribute_gap_to = numpy.argsort(partition_sizes) >= (num_parts - gap)
    # Randomly distribute it
    distribute_gap_to = numpy.concatenate(( numpy.ones(gap, dtype=bool), numpy.zeros(num_parts-gap, dtype=bool) ))
    rng.shuffle(distribute_gap_to)
    partition_sizes[distribute_gap_to] += 1
    assert partition_sizes.sum() == num_items
    indices = numpy.arange(num_items)
    rng.shuffle(indices)
    index        = 0
    for part_index, part_size in enumerate(partition_sizes):
      for i in xrange(part_size):
        doc_index = indices[index]
        partition_map[doc_index, part_index] = True
        index += 1
  return partition_map


def ceil_half(x):
  return x[:int(math.ceil(len(x) / 2.0))]
      



# -*- coding: utf-8 -*-
# Static constants for dealing with different langauge wikis

# List of wikipedia prefixes - not exhaustive
prefixes = ['ru', 'nap', 'ja', 'ta', 'tl', 'az', 'ht', 'et', 'ca', 'wa', 'pms', 
'de', 'nl', 'su', 'bpy', 'fa', 'tr', 'en', 'eu', 'io', 'jv', 'nds', 'pt', 'zh', 
'bn', 'lt', 'la', 'nn', 'no', 'he', 'sh', 'lb', 'ro', 'hr', 'scn', 'da', 'it', 
'te', 'br', 'an', 'fr', 'cs', 'is', 'new', 'lv', 'ast', 'uk', 'mk', 'oc', 'ku', 
'sl', 'sk', 'af', 'ka', 'pl', 'ceb', 'th', 'sv', 'id', 'bs', 'hi', 'el', 'cy', 
'gl', 'vi', 'sq', 'fi', 'mr', 'hu', 'ar', 'bg', 'ms', 'ko', 'be']

# Keyword for identifying category labels in different languages
# TODO: Complete this list
category_identifier =\
  { 'en' : 'Category'
  , 'fr' : 'Catégorie'
  , 'de' : 'Kategorie'
  , 'ja' : 'Category'
  , 'it' : 'Categoria'
  , 'ru' : 'Категория'
  , 'pl' : 'Kategoria'
  , 'nl' : 'Categorie'
  , 'sv' : 'Kategori'
  , 'zh' : 'Category'
  }




import os
import re
import logging

from collections import defaultdict

from config import config
import model
import regexps

logger = logging.getLogger('wikidump.utils')

xml_path = config.get('paths','xml_dumps')

all_prefixes =  [    regexps.dumpfile_name.match(f).group(1) 
                for  f 
                in   os.listdir(xml_path) 
                if   regexps.dumpfile_name.match(f)
                ]

def load_dumps(langs=None, dump_path=None, build_index=False):
  "Load the dumps, and take note of their size"
  # Take note that we will end up loading all the dumps we have for a given language
  # but only returning the last one we find. 
  # TODO: Handle different-dated dumps of the same language
  if langs is None:
    langs = all_prefixes
  if dump_path is None:
    dump_path = xml_path
  dumps = {}
  for path in os.listdir(dump_path):
    if not re.match('|'.join(langs), path): continue
    full_path = os.path.join(dump_path, path)
    d = model.Dump(full_path, build_index=build_index)
    prefix = d.get_dumpfile_prefix() 
    dumps[prefix] = d
  return dumps

def category_map(dump):
  "Compute the category mapping of a particular dump"
  cat_count = defaultdict(list)
  for i in xrange(dump.metadata['size']):
    p = dump.get_page_by_index(i)
    for c in p.categories():
      cat_count[c].append(i)
  return cat_count



from setuptools import setup, find_packages
import sys, os

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()
NEWS = open(os.path.join(here, 'NEWS.txt')).read()


version = '0.1.2'

install_requires = [
    # List your project dependencies here.
    # For more details, see:
    # http://packages.python.org/distribute/setuptools.html#declaring-dependencies
]


setup(name='wikidump',
    version=version,
    description="Tools to manipulate and extract data from wikipedia dumps",
    long_description=README + '\n\n' + NEWS,
    classifiers=[
      # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      "Development Status :: 3 - Alpha"
    , "Environment :: Console"
    , "Intended Audience :: Science/Research"
    , "License :: OSI Approved :: GNU General Public License (GPL)"
    , "Operating System :: OS Independent"
    , "Programming Language :: Python"
    , "Topic :: Scientific/Engineering :: Information Analysis"
    ],
    keywords='wikipedia',
    author='Marco Lui',
    author_email='saffsd@gmail.com',
    url='http://github.com/saffsd/wikidump',
    license='GPL3',
    packages=find_packages('src'),
    package_dir = {'': 'src'},include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    entry_points={
        'console_scripts':
            ['wikidump=wikidump:main']
    }
)



import sys
import csv
import os.path
import logging
import optparse
logging.basicConfig(level=logging.INFO)

logger = logging.getLogger('wikidump')

from wikidump.utils import load_dumps

def main():
  if len(sys.argv) < 2:
    logger.error("Must specify a command!")
    sys.exit(-1)

  command = sys.argv[1]

  if command == 'help':
    # Display a help message
    logger.info("The following commands are supported:")
    logger.info("    index             : builds indexes - warning: can take many hours")
    logger.info("    stats             : print statistics in a tab-delimited CSV format")
    logger.info("    categories [lang] : list categories for 'lang', and the number of documents in each")

  elif command == 'index':
    # Build indices
    load_dumps(build_index=True)

  elif command == 'stats':
    # Display statistics
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.ERROR)

    dumps = load_dumps()
    sizes = dict((d.get_dumpfile_prefix(), d.metadata['size']) for d in dumps.values())

    fields = ['lang', 'filename', 'pages', 'categories']
    outfile = csv.DictWriter(sys.stdout, fields, delimiter='\t')
    for p in sorted(sizes, key=sizes.get, reverse=True):
      d = dict\
            ( lang=p
            , filename= os.path.basename(dumps[p].xml_path)
            , pages=sizes[p]
            , categories=len(dumps[p].categories)
            )
      outfile.writerow(d)

  elif command == 'categories':
    # Dump category distribution
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.ERROR)

    parser = optparse.OptionParser()
    parser.add_option("-l", "--language", dest="lang", help="Relevant language prefix")
    options, args = parser.parse_args(sys.argv[2:])

    dump = load_dumps([options.lang], build_index=True)[options.lang]
    cats = dump.categories

    for c in sorted(cats, key=lambda x:len(cats[x]), reverse=True):
      print "%-4d %s" % (len(cats[c]), c)

  else:
    logging.error("Unknown command: %s", command)
    logging.info("Try the 'help' command.")



from common import shelved
from utils import all_prefixes, raw_doc_lengths

@shelved("docs_under_thresh")
def docs_under_thresh(prefix, thresh):
  doc_lengths = raw_doc_lengths(prefix).values()
  return len(filter(lambda x: x < thresh, doc_lengths))

@shelved("indices_under_thresh")
def indices_under_thresh(prefix, thresh):
  doc_lengths = raw_doc_lengths(prefix).iteritems()
  return [ id for (id,len) in doc_lengths if len < thresh ]

def docids_under_thresh(prefixes, thresh):
  for prefix in prefixes:
    indices = indices_under_thresh(prefix, thresh)
    for index in indices:
      yield "_".join((prefix,str(index)))

